from typing import Dict, Iterable, List, Optional, Tuple, Union, cast

from ..cache_operation_types import CacheGetStatus
from ._utilities._serialization import _bytes_dict_to_string_dict, _bytes_to_string

# Dictionary related responses
DictionaryKey = Union[str, bytes]
DictionaryValue = Union[str, bytes]
StringDictionary = Dict[str, str]
BytesDictionary = Dict[bytes, bytes]
Dictionary = Union[StringDictionary, BytesDictionary]


class CacheDictionaryGetUnaryResponse:
    def __init__(self, value: Optional[DictionaryValue], status: CacheGetStatus):
        self._value = value
        self._status = status

    def value(self) -> Optional[str]:
        if self.status() == CacheGetStatus.MISS:
            return None
        return _bytes_to_string(cast(bytes, self._value))

    def value_as_bytes(self) -> Optional[bytes]:
        if self.status() == CacheGetStatus.MISS:
            return None
        return cast(bytes, self._value)

    def status(self) -> CacheGetStatus:
        return self._status

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, CacheDictionaryGetUnaryResponse)
            and self._value == other._value
            and self._status == other._status
        )

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"CacheDictionaryGetUnaryResponse(value={self._value!r}, status={self._status!r})"


class CacheDictionaryGetMultiResponse:
    def __init__(self, values: List[Optional[DictionaryValue]], status: List[CacheGetStatus]):
        self._values = values
        self._status = status

    def to_list(self) -> List[CacheDictionaryGetUnaryResponse]:
        return [CacheDictionaryGetUnaryResponse(value, status) for value, status in zip(self._values, self._status)]

    def values(self) -> List[Optional[str]]:
        return [
            _bytes_to_string(cast(bytes, value)) if status == CacheGetStatus.HIT else None
            for value, status in zip(self._values, self._status)
        ]

    def values_as_bytes(self) -> List[Optional[bytes]]:
        return [
            cast(bytes, value) if status == CacheGetStatus.HIT else None
            for value, status in zip(self._values, self._status)
        ]

    def status(self) -> List[CacheGetStatus]:
        return self._status

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"CacheDictionaryGetMultiResponse(values={self._values!r}, status={self._status!r})"


class CacheDictionarySetUnaryResponse:
    def __init__(self, dictionary_name: str, key: bytes, value: bytes):
        self._dictionary_name = dictionary_name
        self._key = key
        self._value = value

    def dictionary_name(self) -> str:
        return self._dictionary_name

    def key(self) -> str:
        return _bytes_to_string(self._key)

    def key_as_bytes(self) -> bytes:
        return self._key

    def value(self) -> str:
        return _bytes_to_string(self._value)

    def value_as_bytes(self) -> bytes:
        return self._value

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return (
            f"CacheDictionarySetUnaryResponse(dictionary_name={self._dictionary_name!r}, "
            f"key={self._key!r}, value={self._value!r})"
        )


class CacheDictionarySetMultiResponse:
    def __init__(self, dictionary_name: str, dictionary: BytesDictionary):
        self._dictionary_name = dictionary_name
        self._dictionary = dictionary

    def dictionary_name(self) -> str:
        return self._dictionary_name

    def dictionary(self) -> StringDictionary:
        return _bytes_dict_to_string_dict(self._dictionary)

    def dictionary_as_bytes(self) -> BytesDictionary:
        return self._dictionary

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return (
            f"CacheDictionarySetMultiResponse(dictionary_name={self._dictionary_name!r}, "
            f"dictionary={self._dictionary!r})"
        )


class CacheDictionaryGetAllResponse:
    def __init__(self, value: Optional[BytesDictionary], status: CacheGetStatus):
        self._value = value
        self._status = status

    def value(self) -> Optional[StringDictionary]:
        """Get the dictionary as stored in the cache.

        Returns:
            Optional[StringDictionary]: The dictionary if the cache get was a hit, else None.
        """
        if self.status() != CacheGetStatus.HIT:
            return None

        return _bytes_dict_to_string_dict(cast(BytesDictionary, self._value))

    def value_as_bytes(self) -> Optional[BytesDictionary]:
        """Get the dictionary as stored in the cache.

        Returns:
            Optional[BytesDictionary]: The dictionary if the cache get was a hit, else None.
        """
        if self.status() != CacheGetStatus.HIT:
            return None

        return self._value

    def status(self) -> CacheGetStatus:
        return self._status

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"CacheDictionaryGetAllResponse(value={self._value!r}, status={self._status!r})"


class CacheExistsResponse:
    def __init__(self, keys: Tuple[Union[str, bytes], ...], results: List[bool]):
        self._keys = keys
        self._results = results

    def all(self) -> bool:
        """Test if all the keys exist.

        Examples:
        >>> if not client.exists("my-cache", "key1", "key2", "key3").all():
        ...     log.info("Some key(s) are missing.")
        >>> response = client.exists("my-cache", *keys)
        >>> if response.all()
        ...     log.info("All the keys exist")

        Returns:
            bool: True if all the keys exist. False is at least one does not.
        """
        return all(self._results)

    def num_exists(self) -> int:
        """Count the number of keys that exist.

        Example:
        >>> response = client.exists("my-cache", *keys))
        >>> if response.num_exists() != len(keys):
        ...     log.info("Missing some keys")

        Returns:
            int: The number of keys that exist.
        """
        return sum(self._results)

    def results(self) -> List[bool]:
        """Get a boolean array where each element indicates if the
        corresponding input key exists.

        Example:
        >>> repsonse = client.exists("my-cache", *keys)
        >>> if not response.all():
        ...     mask = response.results()
        ...     missing = [key for key, does_exist in zip(keys, mask) if not does_exist]
        ...     log.info(f"The following keys are missing: {missing}")

        Note the above example illustrates using the existence mask.
        Alternatively one could invoke `missing_keys` to get the missing keys.

        Returns:
            List[bool]: The existence mask.
        """
        return self._results

    def missing_keys(self) -> List[Union[str, bytes]]:
        """List the keys that do not exist.

        Returns:
            List[Union[str, bytes]]: List of queried keys that do not exist.
        """
        return [key for key, does_exist in zip(self._keys, self._results) if not does_exist]

    def present_keys(self) -> List[Union[str, bytes]]:
        """List the keys that exist.

        Returns:
            List[Union[str, bytes]]: List of queried keys that exist.
        """
        return [key for key, does_exist in zip(self._keys, self._results) if does_exist]

    def zip_keys_and_results(self) -> Iterable[Tuple[Union[str, bytes], bool]]:
        """Zip the keys and existence mask.

        Example:
        >>> for key, status in client.exists("my-cache", *keys).zip_keys_and_results():
        ...     print(f"{=key} {=status}")

        Returns:
            Iterable[Tuple[Union[str, bytes], bool]]: An iterator of tuples `(key, exist)`
                where `exist` indicates whether the key exists or not.
        """
        return zip(self._keys, self._results)

    def __bool__(self) -> bool:
        """Test if all the keys exist.

        This is helpful for using the response object in an if-expression:

        >>> if client.exists("my-cache", "key"):
        ...     # Business logic

        Returns:
            bool: True if all the keys exist else False
        """
        return self.all()

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"CacheExistsResponse(keys={self._keys!r}, results={self._results!r})"



import pickle
from typing import cast

from ..._utilities._data_validation import _as_bytes
from ..cache_operation_types import BytesDictionary, Dictionary


def convert_dict_items_to_bytes(dictionary: Dictionary) -> BytesDictionary:
    return {
        _as_bytes(k, "Unsupported type for key: "): _as_bytes(v, "Unsupported type for value: ")
        for k, v in dictionary.items()
    }


def deserialize_dictionary(pickled_dictionary: bytes) -> BytesDictionary:
    return cast(BytesDictionary, pickle.loads(pickled_dictionary))


def serialize_dictionary(dictionary: BytesDictionary) -> bytes:
    return pickle.dumps(dictionary)



import warnings
from typing import List, Optional, Union, cast

from ..._utilities._data_validation import _as_bytes
from ...aio.simple_cache_client import SimpleCacheClient
from .. import INCUBATING_WARNING_MSG
from ..cache_operation_types import (
    BytesDictionary,
    CacheDictionaryGetAllResponse,
    CacheDictionaryGetMultiResponse,
    CacheDictionaryGetUnaryResponse,
    CacheDictionarySetMultiResponse,
    CacheDictionarySetUnaryResponse,
    CacheExistsResponse,
    CacheGetStatus,
    Dictionary,
    DictionaryKey,
    DictionaryValue,
)
from .utils import (
    convert_dict_items_to_bytes,
    deserialize_dictionary,
    serialize_dictionary,
)


class SimpleCacheClientIncubating(SimpleCacheClient):
    def __init__(
        self,
        auth_token: str,
        default_ttl_seconds: int,
        request_timeout_ms: Optional[int] = None,
    ):
        """Creates a SimpleCacheClientIncubating.
        !! Includes non-final, experimental features and APIs subject to change  !!

        Args:
            auth_token: Momento Token to authenticate the requests with Simple Cache Service
            default_ttl_seconds: A default Time To Live in seconds for cache objects created by this client. It is
                possible to override this setting when calling the set method.
            request_timeout_ms: An optional timeout in milliseconds to allow for Get and Set operations to complete.
                Defaults to None, in which case 5 seconds is used. The request will be terminated if it takes longer
                than this value and will result in TimeoutError.
        Raises:
            IllegalArgumentError: If method arguments fail validations
        """
        warnings.warn(INCUBATING_WARNING_MSG)
        super().__init__(auth_token, default_ttl_seconds, request_timeout_ms)

    async def dictionary_set(
        self,
        cache_name: str,
        dictionary_name: str,
        key: DictionaryKey,
        value: DictionaryValue,
        ttl_seconds: Optional[int] = None,
        *,
        refresh_ttl: bool,
    ) -> CacheDictionarySetUnaryResponse:
        """Store a dictionary item in the cache.

        Inserts a `value` for `key` in into a dictionary `dictionary_name`.
        Updates (overwrites) values if the key already exists.

        Args:
            cache_name (str): Name of the cache to store the dictionary in.
            dictionary_name (str): The name of the dictionary in the cache.
            key (DictionaryKey): The key to set.
            value (DictionaryValue): The value to store.
            ttl_seconds (Optional[int], optional): Time to live in seconds for the dictionary
                as a whole.
            refresh_ttl (bool): If, when performing an update, to refresh the ttl.

        Returns:
            CacheDictionarySetUnaryResponse: data stored in the cache
        """
        dictionary_get_response = await self.get(cache_name, dictionary_name)
        cached_dictionary: BytesDictionary = {}
        if dictionary_get_response.status() == CacheGetStatus.HIT:
            cached_dictionary = deserialize_dictionary(cast(bytes, dictionary_get_response.value_as_bytes()))

        bytes_key = _as_bytes(key)
        bytes_value = _as_bytes(value)
        cached_dictionary[bytes_key] = bytes_value

        await self.set(cache_name, dictionary_name, serialize_dictionary(cached_dictionary))
        return CacheDictionarySetUnaryResponse(dictionary_name=dictionary_name, key=bytes_key, value=bytes_value)

    async def dictionary_set_multi(
        self,
        cache_name: str,
        dictionary_name: str,
        dictionary: Dictionary,
        ttl_seconds: Optional[int] = None,
        *,
        refresh_ttl: bool,
    ) -> CacheDictionarySetMultiResponse:
        """Store dictionary items (key-value pairs) in the cache.

        Inserts items from `dictionary` into a dictionary `dictionary_name`.
        Updates (overwrites) values if the key already exists.

        Args:
            cache_name (str): Name of the cache to store the dictionary in.
            dictionary_name (str): The name of the dictionary in the cache.
            dictionary (Dictionary): The items (key-value pairs) to be stored.
            ttl_seconds (Optional[int], optional): Time to live in seconds for the dictionary
                as a whole.
            refresh_ttl (bool): If, when performing an update, to refresh the ttl.

        Returns:
            CacheDictionarySetMultiResponse: data stored in the cache
        """
        dictionary_get_response = await self.get(cache_name, dictionary_name)
        cached_dictionary: BytesDictionary = {}
        if dictionary_get_response.status() == CacheGetStatus.HIT:
            cached_dictionary = deserialize_dictionary(cast(bytes, dictionary_get_response.value_as_bytes()))

        bytes_dictionary = convert_dict_items_to_bytes(dictionary)
        cached_dictionary.update(bytes_dictionary)

        await self.set(cache_name, dictionary_name, serialize_dictionary(cached_dictionary))
        return CacheDictionarySetMultiResponse(dictionary_name=dictionary_name, dictionary=bytes_dictionary)

    async def dictionary_get(
        self,
        cache_name: str,
        dictionary_name: str,
        key: DictionaryKey,
    ) -> CacheDictionaryGetUnaryResponse:
        """Retrieve a dictionary value from the cache.

        Args:
            cache_name (str): Name of the cache to get the dictionary from.
            dictionary_name (str): Name of the dictionary to query.
            key (DictionaryKey): The key to index in the dictionary.

        Returns:
            CacheDictionaryGetUnaryResponse: A wrapper for the value (if present)
                and status (HIT or MISS).
        """
        dictionary_get_response = await self.get(cache_name, dictionary_name)
        if dictionary_get_response.status() == CacheGetStatus.MISS:
            return CacheDictionaryGetUnaryResponse(value=None, status=CacheGetStatus.MISS)

        dictionary: BytesDictionary = deserialize_dictionary(cast(bytes, dictionary_get_response.value_as_bytes()))

        try:
            value = dictionary[_as_bytes(key, "Unsupported type for key: ")]
        except KeyError:
            return CacheDictionaryGetUnaryResponse(value=None, status=CacheGetStatus.MISS)

        return CacheDictionaryGetUnaryResponse(value=value, status=CacheGetStatus.HIT)

    async def dictionary_get_multi(
        self,
        cache_name: str,
        dictionary_name: str,
        *keys: DictionaryKey,
    ) -> CacheDictionaryGetMultiResponse:
        """Retrieve dictionary values from the cache.

        Args:
            cache_name (str): Name of the cache to get the dictionary from.
            dictionary_name (str): Name of the dictionary to query.
            keys (DictionaryKey): The item(s) to index in the dictionary.

        Returns:
            CacheDictionaryGetMultiResponse: a wrapper over a list of values
                and statuses.
        """
        if len(keys) == 0:
            raise ValueError("Argument keys must be non-empty")

        dictionary_get_response = await self.get(cache_name, dictionary_name)
        if dictionary_get_response.status() == CacheGetStatus.MISS:
            return CacheDictionaryGetMultiResponse(
                values=[None for _ in range(len(keys))],
                status=[CacheGetStatus.MISS for _ in range(len(keys))],
            )

        dictionary: BytesDictionary = deserialize_dictionary(cast(bytes, dictionary_get_response.value_as_bytes()))

        values: List[Optional[DictionaryValue]] = []
        results: List[CacheGetStatus] = []
        for key in keys:
            try:
                value = dictionary[_as_bytes(key, "Unsupported type for key: ")]
                values.append(value)
                results.append(CacheGetStatus.HIT)
            except KeyError:
                values.append(None)
                results.append(CacheGetStatus.MISS)

        return CacheDictionaryGetMultiResponse(values=values, status=results)

    async def dictionary_get_all(self, cache_name: str, dictionary_name: str) -> CacheDictionaryGetAllResponse:
        """Retrieve the entire dictionary from the cache.

        Args:
            cache_name (str): Name of the cache to get the dictionary from.
            dictionary_name (str): Name of the dictionary to retrieve.

        Returns:
            CacheDictionaryGetAllResponse: Value (the mapping, if present) and status (HIT or MISS).
        """
        get_response = await self.get(cache_name, dictionary_name)
        if get_response.status() == CacheGetStatus.MISS:
            return CacheDictionaryGetAllResponse(value=None, status=CacheGetStatus.MISS)

        value = deserialize_dictionary(cast(bytes, get_response.value_as_bytes()))
        return CacheDictionaryGetAllResponse(value=value, status=CacheGetStatus.HIT)

    async def exists(self, cache_name: str, *keys: Union[str, bytes]) -> CacheExistsResponse:
        """Test if `keys` exist in the cache.

        Args:
            cache_name (str): Name of the cache to query for the keys.
            keys (str): Key(s) to test for existence.

        Examples:
        >>> if client.exists("my-cache", "key"):
        ...     print("key is present")

        Examples:
        >>> if client.exists("my-cache", "key"):
        ...     print("key is present")

        >>> keys = ["key1", "key2", "key3"]
        ... response = client.exists("my-cache", *keys)
        ... if response.all():
        ...     print("All keys are present")
        ... else:
        ...     missing_keys = response.missing_keys()
        ...     print(f"num_missing={len(missing_keys)}; num_exists={response.num_exists()}")
        ...     print(f"{=missing_keys}")

        Returns:
            CacheExistsResponse: Wrapper object containing the results
                of the exists test.
        """
        get_multi_response = await self.get_multi(cache_name, *keys)
        mask = [status == CacheGetStatus.HIT for status in get_multi_response.status()]
        return CacheExistsResponse(keys, mask)



import asyncio
from typing import Mapping, Optional, Union

from momento_wire_types.cacheclient_pb2 import _DeleteRequest, _GetRequest, _SetRequest

from .. import _cache_service_errors_converter
from .. import cache_operation_types as cache_sdk_ops
from .. import logs
from .._utilities._data_validation import (
    _as_bytes,
    _make_metadata,
    _validate_cache_name,
    _validate_ttl,
)
from . import _scs_grpc_manager

_DEFAULT_DEADLINE_SECONDS = 5.0  # 5 seconds


class _ScsDataClient:
    """Internal"""

    def __init__(
        self,
        auth_token: str,
        endpoint: str,
        default_ttl_seconds: int,
        operation_timeout_ms: Optional[int],
    ):
        self._logger = logs.logger
        self._logger.debug("Simple cache data client instantiated with endpoint: %s", endpoint)
        self._default_deadline_seconds = (
            _DEFAULT_DEADLINE_SECONDS if not operation_timeout_ms else operation_timeout_ms / 1000.0
        )
        self._grpc_manager = _scs_grpc_manager._DataGrpcManager(auth_token, endpoint)
        _validate_ttl(default_ttl_seconds)
        self._default_ttlSeconds = default_ttl_seconds
        self._endpoint = endpoint

    def get_endpoint(self) -> str:
        return self._endpoint

    async def set(
        self,
        cache_name: str,
        key: Union[str, bytes],
        value: Union[str, bytes],
        ttl_seconds: Optional[int],
    ) -> cache_sdk_ops.CacheSetResponse:
        _validate_cache_name(cache_name)
        try:
            self._logger.log(logs.TRACE, "Issuing a set request with key %s", str(key))
            item_ttl_seconds = self._default_ttlSeconds if ttl_seconds is None else ttl_seconds
            _validate_ttl(item_ttl_seconds)
            set_request = _SetRequest()
            set_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            set_request.cache_body = _as_bytes(value, "Unsupported type for value: ")
            set_request.ttl_milliseconds = item_ttl_seconds * 1000
            await self._grpc_manager.async_stub().Set(
                set_request,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._logger.log(logs.TRACE, "Set succeeded for key: %s", str(key))
            return cache_sdk_ops.CacheSetResponse(set_request.cache_key, set_request.cache_body)
        except Exception as e:
            self._logger.log(logs.TRACE, "Set failed for %s with response: %s", str(key), e)
            raise _cache_service_errors_converter.convert(e)

    async def set_multi(
        self,
        cache_name: str,
        items: Union[Mapping[str, str], Mapping[bytes, bytes]],
        ttl_seconds: Optional[int] = None,
    ) -> cache_sdk_ops.CacheSetMultiResponse:
        _validate_cache_name(cache_name)

        try:
            request_promises = [self.set(cache_name, key, value, ttl_seconds) for key, value in items.items()]

            # A note on `return_exceptions=True`: because we're gathering the results,
            # if an individual promise raises an exception, we want the others to finish gracefully.
            responses = await asyncio.gather(
                *request_promises,
                return_exceptions=True,
            )

            for response in responses:
                if isinstance(response, Exception):
                    raise response

            items_as_bytes = {response.key_as_bytes(): response.value_as_bytes() for response in responses}
            return cache_sdk_ops.CacheSetMultiResponse(items=items_as_bytes)
        except Exception as e:
            self._logger.debug("multi-set failed with error: %s", e)
            # re-raise any error caught here is fatal error with overall handling of request objects
            raise _cache_service_errors_converter.convert(e)

    async def get(self, cache_name: str, key: Union[str, bytes]) -> cache_sdk_ops.CacheGetResponse:

        _validate_cache_name(cache_name)
        try:
            self._logger.log(logs.TRACE, "Issuing a get request with key %s", str(key))
            get_request = _GetRequest()
            get_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            response = await self._grpc_manager.async_stub().Get(
                get_request,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._logger.log(logs.TRACE, "Received a get response for %s", str(key))
            return cache_sdk_ops.CacheGetResponse.from_grpc_response(response)
        except Exception as e:
            self._logger.log(logs.TRACE, "Get failed for %s with response: %s", str(key), e)
            raise _cache_service_errors_converter.convert(e)

    async def get_multi(
        self,
        cache_name: str,
        *keys: Union[str, bytes],
    ) -> cache_sdk_ops.CacheGetMultiResponse:
        _validate_cache_name(cache_name)
        try:
            request_promises = [self.get(cache_name, key) for key in keys]

            # A note on `return_exceptions=True`: because we're gathering the results,
            # if an individual promise raises an exception, we want the others to finish gracefully.
            responses = await asyncio.gather(
                *request_promises,
                return_exceptions=True,
            )
            for response in responses:
                if isinstance(response, Exception):
                    raise response
        except Exception as e:
            self._logger.debug("get_multi failed with response: %s", e)
            raise _cache_service_errors_converter.convert(e)

        return cache_sdk_ops.CacheGetMultiResponse(responses=responses)

    async def delete(self, cache_name: str, key: Union[str, bytes]) -> cache_sdk_ops.CacheDeleteResponse:
        _validate_cache_name(cache_name)
        try:
            self._logger.log(logs.TRACE, "Issuing a delete request with key %s", str(key))
            delete_request = _DeleteRequest()
            delete_request.cache_key = _as_bytes(key, "Unsupported type for key: ")
            await self._grpc_manager.async_stub().Delete(
                delete_request,
                metadata=_make_metadata(cache_name),
                timeout=self._default_deadline_seconds,
            )
            self._logger.log(logs.TRACE, "Received a delete response for %s", str(key))
            return cache_sdk_ops.CacheDeleteResponse()
        except Exception as e:
            self._logger.debug("Delete failed for %s with response: %s", str(key), e)
            raise _cache_service_errors_converter.convert(e)

    async def close(self) -> None:
        await self._grpc_manager.close()



import warnings
from typing import Optional, Union

from .._async_utils import wait_for_coroutine
from .._utilities._data_validation import _validate_request_timeout
from ..simple_cache_client import SimpleCacheClient
from . import INCUBATING_WARNING_MSG
from .aio import simple_cache_client as aio
from .cache_operation_types import (
    CacheDictionaryGetAllResponse,
    CacheDictionaryGetMultiResponse,
    CacheDictionaryGetUnaryResponse,
    CacheDictionarySetMultiResponse,
    CacheDictionarySetUnaryResponse,
    CacheExistsResponse,
    Dictionary,
    DictionaryKey,
    DictionaryValue,
)


class SimpleCacheClientIncubating(SimpleCacheClient):
    def __init__(
        self,
        auth_token: str,
        default_ttl_seconds: int,
        request_timeout_ms: Optional[int] = None,
    ):
        """Creates a SimpleCacheClientIncubating.
        !! Includes non-final, experimental features and APIs subject to change  !!

        Args:
            auth_token: Momento Token to authenticate the requests with Simple Cache Service
            default_ttl_seconds: A default Time To Live in seconds for cache objects created by this client. It is
                possible to override this setting when calling the set method.
            request_timeout_ms: An optional timeout in milliseconds to allow for Get and Set operations to complete.
                Defaults to None, in which case 5 seconds is used. The request will be terminated if it takes longer
                than this value and will result in TimeoutError.
        Raises:
            IllegalArgumentError: If method arguments fail validations
        """

        warnings.warn(INCUBATING_WARNING_MSG)
        self._init_loop()
        self._momento_async_client: aio.SimpleCacheClientIncubating = aio.SimpleCacheClientIncubating(
            auth_token=auth_token,
            default_ttl_seconds=default_ttl_seconds,
            request_timeout_ms=request_timeout_ms,
        )

    def dictionary_set(
        self,
        cache_name: str,
        dictionary_name: str,
        key: DictionaryKey,
        value: DictionaryValue,
        ttl_seconds: Optional[int] = None,
        *,
        refresh_ttl: bool,
    ) -> CacheDictionarySetUnaryResponse:
        """Store a dictionary item in the cache.

        Inserts a `value` for `key` in into a dictionary `dictionary_name`.
        Updates (overwrites) values if the key already exists.

        Args:
            cache_name (str): Name of the cache to store the dictionary in.
            dictionary_name (str): The name of the dictionary in the cache.
            key (DictionaryKey): The key to set.
            value (DictionaryValue): The value to store.
            ttl_seconds (Optional[int], optional): Time to live in seconds for the dictionary
                as a whole.
            refresh_ttl (bool): If, when performing an update, to refresh the ttl.

        Returns:
            CacheDictionarySetUnaryResponse: data stored in the cache
        """
        coroutine = self._momento_async_client.dictionary_set(
            cache_name,
            dictionary_name,
            key,
            value,
            ttl_seconds,
            refresh_ttl=refresh_ttl,
        )
        return wait_for_coroutine(self._loop, coroutine)

    def dictionary_set_multi(
        self,
        cache_name: str,
        dictionary_name: str,
        dictionary: Dictionary,
        ttl_seconds: Optional[int] = None,
        *,
        refresh_ttl: bool,
    ) -> CacheDictionarySetMultiResponse:
        """Store dictionary items (key-value pairs) in the cache.

        Inserts items from `dictionary` into a dictionary `dictionary_name`.
        Updates (overwrites) values if the key already exists.

        Args:
            cache_name (str): Name of the cache to store the dictionary in.
            dictionary_name (str): The name of the dictionary in the cache.
            dictionary (Dictionary): The items (key-value pairs) to be stored.
            ttl_seconds (Optional[int], optional): Time to live in seconds for the dictionary
                as a whole.
            refresh_ttl (bool): If, when performing an update, to refresh the ttl.

        Returns:
            CacheDictionarySetMultiResponse: data stored in the cache
        """
        coroutine = self._momento_async_client.dictionary_set_multi(
            cache_name,
            dictionary_name,
            dictionary,
            ttl_seconds,
            refresh_ttl=refresh_ttl,
        )
        return wait_for_coroutine(self._loop, coroutine)

    def dictionary_get(
        self,
        cache_name: str,
        dictionary_name: str,
        key: DictionaryKey,
    ) -> CacheDictionaryGetUnaryResponse:
        """Retrieve a dictionary value from the cache.

        Args:
            cache_name (str): Name of the cache to get the dictionary from.
            dictionary_name (str): Name of the dictionary to query.
            key (DictionaryKey): The key to index in the dictionary.

        Returns:
            CacheDictionaryGetUnaryResponse: A wrapper for the value (if present)
                and status (HIT or MISS).
        """
        coroutine = self._momento_async_client.dictionary_get(cache_name, dictionary_name, key)
        return wait_for_coroutine(self._loop, coroutine)

    def dictionary_get_multi(
        self,
        cache_name: str,
        dictionary_name: str,
        *keys: DictionaryKey,
    ) -> CacheDictionaryGetMultiResponse:
        """Retrieve dictionary values from the cache.

        Args:
            cache_name (str): Name of the cache to get the dictionary from.
            dictionary_name (str): Name of the dictionary to query.
            keys (DictionaryKey): The item(s) to index in the dictionary.

        Returns:
            CacheDictionaryGetMultiResponse: a wrapper over a list of values
                and statuses.
        """
        coroutine = self._momento_async_client.dictionary_get_multi(cache_name, dictionary_name, *keys)
        return wait_for_coroutine(self._loop, coroutine)

    def dictionary_get_all(self, cache_name: str, dictionary_name: str) -> CacheDictionaryGetAllResponse:
        """Retrieve the entire dictionary from the cache.

        Args:
            cache_name (str): Name of the cache to get the dictionary from.
            dictionary_name (str): Name of the dictionary to retrieve.

        Returns:
            CacheDictionaryGetAllResponse: Value (the mapping, if present) and status (HIT or MISS).
        """
        coroutine = self._momento_async_client.dictionary_get_all(cache_name, dictionary_name)
        return wait_for_coroutine(self._loop, coroutine)

    def exists(self, cache_name: str, *keys: Union[str, bytes]) -> CacheExistsResponse:
        """Test if `keys` exist in the cache.

        Args:
            cache_name (str): Name of the cache to query for the keys.
            keys (str): Key(s) to test for existence.

        Examples:
        >>> if client.exists("my-cache", "key"):
        ...     print("key is present")

        >>> keys = ["key1", "key2", "key3"]
        ... response = client.exists("my-cache", *keys)
        ... if response.all():
        ...     print("All keys are present")
        ... else:
        ...     missing_keys = response.missing_keys()
        ...     print(f"num_missing={len(missing_keys)}; num_exists={response.num_exists()}")
        ...     print(f"{=missing_keys}")

        Returns:
            CacheExistsResponse: Wrapper object containing the results
                of the exists test.
        """
        coroutine = self._momento_async_client.exists(cache_name, *keys)
        return wait_for_coroutine(self._loop, coroutine)


def init(
    auth_token: str,
    item_default_ttl_seconds: int,
    request_timeout_ms: Optional[int] = None,
) -> SimpleCacheClientIncubating:
    """Creates a SimpleCacheClientIncubating.
    !! Includes non-final, experimental features and APIs subject to change  !!

    Args:
        auth_token: Momento Token to authenticate the requests with Simple Cache Service
        item_default_ttl_seconds: A default Time To Live in seconds for cache objects created by this client. It is
            possible to override this setting when calling the set method.
        request_timeout_ms: An optional timeout in milliseconds to allow for Get and Set operations to complete.
            Defaults to 5 seconds. The request will be terminated if it takes longer than this value and will result
            in TimeoutError.
    Returns:
        SimpleCacheClientIncubating
    Raises:
        IllegalArgumentError: If method arguments fail validations
    """
    _validate_request_timeout(request_timeout_ms)
    return SimpleCacheClientIncubating(auth_token, item_default_ttl_seconds, request_timeout_ms)



from typing import Dict


def _bytes_to_string(value: bytes) -> str:
    """Decode bytes to a UTF-8 string.

    Args:
        value (bytes): The bytes to decode

    Returns:
        str: UTF-8 representation of bytes

    Raises:
        UnicodeDecodeError
    """
    return value.decode(encoding="utf-8")


def _bytes_dict_to_string_dict(dictionary: Dict[bytes, bytes]) -> Dict[str, str]:
    """Convert byte-byte dictionary to string-string dictionary.

    Args:
        dictionary (Dict[bytes, bytes]): The dictionary to convert.

    Returns:
        Dict[str, str]: The dictionary with decoded items.
    """
    return {_bytes_to_string(k): _bytes_to_string(v) for k, v in dictionary.items()}
